#
# _modules/gallery.pm
#
# Developed by Lubomir Host 'rajo' <rajo AT platon.sk>
# Copyright (c) 2004-2006 Platon Group, http://platon.sk/
# Licensed under terms of GNU General Public License.
# All rights reserved.
#
# Changelog:
# 2005-07-19 - created
#

# $Platon: web-apps/fastcgi-gallery/_modules/gallery.pm,v 1.5 2005/10/18 12:57:19 rajo Exp $

package gallery;

use strict;

use vars qw($VERSION);

$VERSION = "0.1";

use Image::Info qw(image_info);
use Image::Size qw(imgsize);
use Image::Imlib2;
use Text::Template;
use POSIX qw(floor strftime);
use Memoize;
use MIME::Types qw(by_suffix by_mediatype import_mime_types);

use File::stat;
use File::Spec;
use File::Copy;
use File::Basename qw( basename fileparse );
use File::Path qw(mkpath); # equivalent to mkdir
use Platon::File::Spec::Unix;

use Data::Dumper;

sub new()
{ # {{{
    my $this  = shift;
    my $class = ref($this) || $this;
	my $config = shift;
    my $self  = {
		_config		=> $config,
		_memoized	=> 0,
	};

    bless $self, $class;

	if ((not $self->{_memoized}) and ($self->config('GalleryMemoize'))) {
		Memoize::memoize('get_imageinfo');
		$self->{_memoized} = 1;
	}

    return $self;
} # }}}

sub config()
{ # {{{
	my ($self, $key) = @_;
	return $self->{_config}->{$key};
} # }}}

sub error()
{ # {{{
	my ($self, $error, @param) = @_;

	return {
		'error'		=> $error,
		'caller'	=> [caller],
		'param'		=> \@param,
	};
} # }}}

sub handler { # {{{
	my ($self, $query, $cookie) = @_;

	$self->{__use_cache} = 1;
	my %tpl_vars;

	my $uri = $query->{gallery} || '/';
	$uri		= Platon::File::Spec::Unix->canonpath("/$uri");
	$uri		=~ s|^(\.\./)+||g; # ignore leading ../
	$uri		=~ s|^\.$|/|g;	# .  -->  /
	$uri		=~ s|^\./$|/|g;	# ./ -->  /
	$self->{_uri} = $uri;

	my $filename = $query->{gallery} || '/';
	$filename	= Platon::File::Spec::Unix->canonpath("/$filename");
	$filename	=~ s|^(\.\./)+||g; # ignore leading ../
	$filename	=~ s|^\.$|/|g;	# .  -->  /
	$filename	=~ s|^\./$|/|g;	# ./ -->  /
	$self->{_filename} = $filename;

	# Lookup the file in the cache and scale the image if the cached
	# image does not exist
	if ($filename =~ m/^\/?\.cache\//) { # {{{

		$filename		=~ s/^\/?\.cache\///;
		my ($fname, $path, $dummy) = fileparse($filename, ''); # qr{\..*}

		my ($image_width, $image_height) = ($fname =~ m/^(\d+)x(\d+)\-/);

		return $self->error('cache_error') unless (defined($image_width) and defined($image_height));
		
		$fname		=~ s/^(\d+)x(\d+)\-//;
		$filename	= Platon::File::Spec::Unix->canonpath($self->config('GalleryRootPath') . "/$path/$fname");

		return $self->error('not_found', $filename) unless (-f $filename or -d $filename);

		my ($width, $height, $type)	= imgsize($filename);
		my $imageinfo				= $self->get_imageinfo($filename, $type, $width, $height);
		my $cached					= $self->scale_picture($filename, $image_width, $image_height, $imageinfo);
		my $file					= cache_dir($self, 0);

		my ($mime_type, $encoding) = by_suffix($file);
		return {
			use_cache	=> $self->{__use_cache},
			content	=> {
				mime_type	=> $mime_type,
				encoding	=> $encoding,
				filename	=> $file,
				data		=> $self->get_filecontent($file),
			},
		};
	} # }}}

	my $top_location	= ($self->config('GalleryDocLocation') or '/');
	my $topdir			= ($self->config('GalleryRootPath') or '.');
	my $doc_pattern		= ($self->config('GalleryDocFile') or '\.(mpe?g|avi|mov|asf|wmv|doc|mp3|ogg|pdf|rtf|wav|dlt|html?|csv|eps|sxw)$');
	my $img_pattern		= ($self->config('GalleryImgFile') or '\.(jpe?g|png|tiff?|ppm)$');

	my $abs_filename = Platon::File::Spec::Unix->canonpath("$topdir/$filename");

	return $self->error('not_found', $abs_filename) unless (-f $abs_filename or -d $abs_filename);

	# make redirect to file
	if (-f $abs_filename && $abs_filename !~ m/$img_pattern/i) {
		if ($abs_filename =~ m/$doc_pattern/i) {
			return {
				use_cache	=> $self->{__use_cache},
				location	=> "$top_location/$filename",
			};
		}
	}

	if (-d $abs_filename) { # {{{ directory listing

		unless (-d cache_dir($self, 0)) {
			# create cache directories
			mkpath(cache_dir($self, 0));
		}

		# stop slideshow on directory change
		$tpl_vars{cookie}->{GalleryInterval} = 0;

		$tpl_vars{META}		= "";
		$tpl_vars{ICONS}	= $self->config('GalleryIconsURI');
		$tpl_vars{DOC_LOCATION}	= $self->config('GalleryDocLocation');

		opendir(DIR, $abs_filename) or return $self->error('permision_denied');

		$tpl_vars{menu} = $self->generate_menu();

		# Read, sort, and filter files
		my @files = grep { !/^\./ && !/^CVS$/ && -f "$abs_filename/$_" } readdir (DIR);

		@files = gallerysort($self, $abs_filename, @files);

		my @downloadable_files;

		# Remove unwanted files from list # {{{
		my @new_files = ();
		foreach my $picture (@files) {
			if ($picture =~ /$doc_pattern/i) {
				push (@downloadable_files, $picture);
			}
			elsif ($picture =~ /$img_pattern/i) {
				push (@new_files, $picture);
			}
		} # }}}
		@files = @new_files;

		# Read and sort directories
		rewinddir (DIR);
		my @directories = grep { !/^\./ && !/^CVS$/ && -d "$abs_filename/$_" && -r "$abs_filename/$_" } readdir (DIR);
		my $dirsortby = $self->config('GalleryDirSortBy') || $self->config('GallerySortBy') || '';

		if ($dirsortby && $dirsortby =~ m/^(size|atime|mtime|ctime)$/) {
			@directories = map(/^\d+ (.*)/, sort map(stat("$abs_filename/$_")->$dirsortby()." $_", @directories));
		} else {
			@directories = sort @directories;
		}

		closedir(DIR);
		$tpl_vars{LISTING} = {
			directories			=> [ map {
				{
					name	=> $_,
					title	=> get_dirtitle($self, "$abs_filename/$_", $_),
					subentries	=> $self->get_subentries_count($abs_filename, $_),
				}
			} @directories ],
			files				=> [ map {
				get_fileinfo($self, $abs_filename, $_)
			} @files ],
			downloadable_files	=> [ map {
				{
					name	=> $_,
					size	=> stat("$abs_filename/$_")->size,
					type	=> lc([$_ =~ /$doc_pattern/i]->[0]), # return matched file extension
					filetype	=> lc(
						$_ =~ m/\.(mpe?g|avi|mov|asf|wmv)$/i	? "video-$1"		:
						$_ =~ m/\.(txt|html?)$/i				? "text-$1"			:
						$_ =~ m/\.(mp3|ogg|wav)$/i				? "sound-$1"		:
						$_ =~ m/\.(doc|pdf|rtf|csv|eps|sxw)$/i	? "application-$1"	: 'unknown'),
				}
			} @downloadable_files ],
		};

		my $start_at = 1;
		if (defined($query->{GalleryStart})) {
			$start_at = int($query->{GalleryStart});
			if ($start_at < 1) {
				$start_at = 1;
			}
		}

		if (-f "$abs_filename/.comment") {
			my $comment_ref = get_comment("$abs_filename/.comment");
			$tpl_vars{COMMENT} = $comment_ref->{COMMENT} if $comment_ref->{COMMENT};
			$tpl_vars{TITLE} = $comment_ref->{TITLE} if $comment_ref->{TITLE};
		}
		else {
			$tpl_vars{COMMENT} = '';
		}

		$tpl_vars{use_cache} = $self->{__use_cache};

		return \%tpl_vars;

	} # }}} directory listing
	else { # {{{ image page

		my %tpl_vars;
		
		# original size
		if (defined($query->{GalleryShowOriginal}) && $query->{GalleryShowOriginal} eq '1') { # {{{
			if ($self->config('GalleryAllowOriginal') ? 1 : 0) {
				
				my ($mime_type, $encoding) = by_suffix($abs_filename);
				return {
					use_cache	=> $self->{__use_cache},
					content	=> {
						mime_type	=> $mime_type,
						encoding	=> $encoding,
						filename	=> $abs_filename,
						data		=> $self->get_filecontent($abs_filename),
					},
				};
			}
			else {
				return $self->error('permision_denied', $abs_filename);
			}
		} # }}}

		# Create cache dir if not existing
		my ($picfilename, $abs_path, $dummy) = fileparse($abs_filename, ''); # qr{\..*}
		my ($dummy2, $path, $dummy3) = fileparse($filename, ''); # qr{\..*}
		my $cache_path = cache_dir($self, 1);

		unless (-d cache_dir($self, 1)) {
			# create cache directories
			mkpath(cache_dir($self, 1));
		}

		my ($orig_width, $orig_height, $type) = imgsize($abs_filename);

		my $imageinfo = get_imageinfo($self, $abs_filename, $type, $orig_width, $orig_height);

		my $original_size = $orig_height;
 		if ($orig_width > $orig_height) {
			$original_size = $orig_width;
 		}

		# Check if the selected width is allowed
		my $sizes = $self->config('GallerySizes') || [ qw(320 640 800 1024 1600) ];
		my $width = $query->{GallerySize} || $cookie->{GallerySize} || $sizes->[0];
		unless ((grep $width == $_, @$sizes) or ($width == $original_size)) {
			#return $self->error('invalid_width');
			$width = $sizes->[0]; # invalid width, continue with smallest value
		}
		$tpl_vars{cookie}->{GallerySize} = $width;

		my $scale;
		my $image_width;
		if ($orig_width < $orig_height) {
			$scale = ($orig_height ? $width / $orig_height : 1);
			$image_width = $width * $orig_width / $orig_height;
		}
		else {
			$scale = ($orig_width ? $width / $orig_width : 1);
			$image_width = $width;
		}

		my $height = $orig_height * $scale;

		$image_width = floor($image_width);
		$width       = floor($width);
		$height      = floor($height);

		my $cached = get_scaled_picture_name($self, $filename, $type, $image_width, $height);
		
		my $resolution = (($image_width > $orig_width) && ($height > $orig_height)) ? 
			"$orig_width x $orig_height" : "$image_width x $height";

		$tpl_vars{cur_image} = {
			name		=> $picfilename,
			url			=> "$path$picfilename",
			src			=> get_cache_filename($self, $path, $cached),
			meta		=> '',
			resolution	=> $resolution,
			width		=> (($image_width > $orig_width) && ($height > $orig_height)) ? $orig_width		: $image_width,
			height		=> (($image_width > $orig_width) && ($height > $orig_height)) ? $orig_height	: $height,
			info		=> $imageinfo,
		};
		$tpl_vars{cur_image}->{info}->{orig_imagesize}	= -s $abs_filename;
		$tpl_vars{image_sizes}	= $sizes;
		$tpl_vars{menu} = generate_menu($self);
			
	
		my $exif_mode = $self->config('GalleryEXIFMode') || 'namevalue';

		unless (opendir(DATADIR, $abs_path)) {
			return $self->error('permision_denied', $abs_filename);
		}
		my @pictures = grep { /$img_pattern/i } readdir (DATADIR);
		closedir(DATADIR);
		@pictures = gallerysort($self, $abs_path, @pictures);

		$tpl_vars{total} = scalar @pictures;

		my $prevpicture;
		my $nextpicture;
	
		# get index number for current image
		for (my $i = 0; $i < scalar(@pictures); $i++) { # {{{
			if ($pictures[$i] eq $picfilename) {

				$tpl_vars{number} = $i + 1;

				$prevpicture = $pictures[$i - 1];
				$nextpicture = $pictures[$i + 1];
				my $displayprev = ($i > 0 ? 1 : 0);

				if ($self->config("GalleryWrapNavigation")) {
					$prevpicture = $pictures[$i > 0 ? $i - 1 : scalar(@pictures)];
					$nextpicture = $pictures[$i == scalar(@pictures) ? 0 : $i + 1];
					$displayprev = 1;
				}	

				if ($prevpicture and $displayprev) {
					my ($orig_width, $orig_height, $type) = imgsize($abs_path.$prevpicture);
					my ($thumbnailwidth, $thumbnailheight) = get_thumbnailsize($self, $orig_width, $orig_height);	
					my $imageinfo = get_imageinfo($self, $path.$prevpicture, $type, $orig_width, $orig_height);
					my $cached = get_scaled_picture_name($self, $path.$prevpicture, $type, $thumbnailwidth, $thumbnailheight);
					$tpl_vars{prev_image} = {
						name	=> $prevpicture,
						url		=> "$path$prevpicture",
						width	=> $thumbnailwidth,
						height	=> $thumbnailheight,
						src		=> get_cache_filename($self, $path, $cached),
					};
				}
				else {
					$tpl_vars{prev_image} = { };
				}

				if ($nextpicture) {
					my ($orig_width, $orig_height, $type) = imgsize($abs_path.$nextpicture);
					my ($thumbnailwidth, $thumbnailheight) = get_thumbnailsize($self, $orig_width, $orig_height);	
					my $imageinfo = get_imageinfo($self, $path.$nextpicture, $type, $thumbnailwidth, $thumbnailheight);
					my $cached = get_scaled_picture_name($self, $path.$nextpicture, $type, $thumbnailwidth, $thumbnailheight);
					$tpl_vars{next_image} = {
						name	=> $nextpicture,
						url		=> "$path$nextpicture",
						width	=> $thumbnailwidth,
						height	=> $thumbnailheight,
						src		=> ".cache$path$cached",
						src		=> get_cache_filename($self, $path, $cached),
						# Mozilla prefetch feature: http://www.mozilla.org/projects/netlib/Link_Prefetching_FAQ.html
						prefetch	=> get_cache_filename($self, $path, get_scaled_picture_name($self, $path.$nextpicture, $type, $image_width, $height)),
					};
				}
				else {
					$tpl_vars{next_image} = { };
				}
			}
		} # }}}

		if (-f $abs_path . $picfilename . '.comment') {
			my $comment_ref = get_comment($abs_path . $picfilename . '.comment');
			$tpl_vars{COMMENT} = $comment_ref->{COMMENT} if $comment_ref->{COMMENT};
			$tpl_vars{TITLE} = $comment_ref->{TITLE} if $comment_ref->{TITLE};
		} else {
			$tpl_vars{COMMENT} = '';
		}


		my $scaleable = 0;
		foreach my $size (@$sizes) {
			if ($size<=$original_size) {
				my %sizes_vars;
				#$sizes_vars{IMAGEURI} = uri_escape($r->uri(), $escape_rule);
				$sizes_vars{SIZE}     = $size;
				$sizes_vars{WIDTH}    = $size;
				if ($width == $size) {
					#$tpl_vars{SIZES} .= $templates{scaleactive}->fill_in(HASH => \%sizes_vars);
				}
				else {
					#$tpl_vars{SIZES} .= $templates{scale}->fill_in(HASH => \%sizes_vars);
					$scaleable = 1;
				}
			}
		}

		unless ($scaleable) {
			my %sizes_vars;
			#$sizes_vars{IMAGEURI} = uri_escape($r->uri(), $escape_rule);
			$sizes_vars{SIZE}     = $original_size;
			$sizes_vars{WIDTH}    = $original_size;
			#$tpl_vars{SIZES} .= $templates{scaleactive}->fill_in(HASH => \%sizes_vars);
		}


		my $slideshow_intervals = $self->config('GallerySlideshowIntervals') || [ qw(3 5 10 15 30) ];
		my $interval = $query->{GalleryInterval} || $cookie->{GalleryInterval} || 0;
		unless ((grep $interval == $_, @$slideshow_intervals)) {
			#return $self->error('invalid_interval');
			$interval = 0; # invalid interval, stop slideshow
		}
		$tpl_vars{cookie}->{GalleryInterval} = $interval;
		$tpl_vars{slideshow_intervals} = $slideshow_intervals;

		if ($interval and $nextpicture) {
			$tpl_vars{refresh} = {
				interval	=> $interval,
				url			=> "$path$nextpicture",
			};
		}

		$tpl_vars{use_cache} = $self->{__use_cache};

		return \%tpl_vars;
		
	} # }}} image page

} # }}}

sub cache_dir { # {{{

	my ($self, $strip_filename) = @_;

	my $cache_root = $self->config('GalleryCacheDir');
	$cache_root ||= '/var/tmp/Apache-Gallery/';

	# If the uri contains .cache we need to remove it
	my $uri = $self->{_filename};
	$uri =~ s/^\/?\.cache\///;

	my (undef, $dirs, $filename) = File::Spec->splitpath($uri);
	# We don't need a volume as this is a relative path

	if ($strip_filename) {
		return(File::Spec->canonpath(File::Spec->catdir($cache_root, $dirs)));
	} else {
		return(File::Spec->canonpath(File::Spec->catfile($cache_root, $dirs, $filename)));
	}
} # }}}

sub get_cache_filename { # {{{

	my ($self, $path, $filename) = @_;
	my $rv;
	
	if ($self->config('GalleryDirectCacheAccess')) {
		if (-f $self->config('GalleryCacheDir') . "$path$filename") {
			$rv = {
				url	=> $self->config('GalleryDirectCacheAccess') . "$path$filename",
				direct_cache	=> 1,
			};
		}
		else {
			$rv = {
				url	=> ".cache$path$filename",
				direct_cache	=> 0,
			};
			$self->{__use_cache} &&= 0; # don't cache this request
		}
	}
	else {
		$rv = {
			url	=> ".cache$path$filename",
			direct_cache	=> 0,
		};
	}

	return $rv;

} # }}}

sub get_scaled_picture_name { # {{{

	my ($self, $fullpath, $type, $width, $height) = @_;

	my @dirs = split(/\//, $fullpath);
	my $filename = pop(@dirs);
	my $newfilename;

	if ($type eq 'JPG') {
		$newfilename = $width . 'x' . $height . '-' . $filename;
	} else {
		$newfilename = $width . 'x' . $height . '-' . $filename;
		# needs to be configurable
		$newfilename =~ s/\.(\w+)$/-$1\.jpg/;
	}

	return $newfilename;
	
} # }}}

sub get_thumbnailsize { # {{{
	my ($self, $orig_width, $orig_height) = @_;

	my $gallerythumbnailsize = $self->config('GalleryThumbnailSize');
	$gallerythumbnailsize ||= '100x75';

	my ($thumbnailwidth, $thumbnailheight) = split(/x/i, $gallerythumbnailsize);

	my $width = int($thumbnailwidth) || 100;
	my $height = int($thumbnailheight) || 75;

	# If the image is rotated, flip everything around.
	if (defined $self->config('GalleryThumbnailSizeLS')
			and $self->config('GalleryThumbnailSizeLS') eq '1'
			and $orig_width < $orig_height) {

		$width = $thumbnailheight;
		$height = $thumbnailwidth;
	}

	my $scale = ($orig_width ? $width/$orig_width : 1);

	if ($orig_height) {
		if ($orig_height * $scale > $thumbnailheight) {
			$scale = $height/$orig_height;
			$width = $orig_width * $scale;
		}
	}

	$height = $orig_height * $scale;

	$height = floor($height);
	$width  = floor($width);

	return ($width, $height);
} # }}}

sub get_fileinfo($$$;) { # {{{
	my ($self, $path, $filename) = @_;
	my $file = "$path/$filename";
	my ($width, $height, $type);  
	my @filetypes = qw(JPG TIF PNG PPM GIF);
	my $img_pattern		= ($self->config('GalleryImgFile') or '\.(jpe?g|png|tiff?|ppm)$');
	
	($width, $height, $type) = imgsize($file);

	return { } if $type eq 'Data stream is not a known image file format';

	return unless (grep $type eq $_, @filetypes);

	my ($thumbnailwidth, $thumbnailheight) = get_thumbnailsize($self, $width, $height);	
	my $imageinfo = get_imageinfo($self, $file, $type, $width, $height);
	my $cached = get_scaled_picture_name($self, $filename, $type, $thumbnailwidth, $thumbnailheight);

	my $rotate = readfile_getnum($self, $imageinfo, "$file.rotate");

	my $info = {
		name	=> $filename,
		size	=> stat($file)->size,
		type	=> lc([$file =~ /$img_pattern/i]->[0]), # return matched file extension
		date	=> ($imageinfo->{DateTimeOriginal} or strftime("%Y-%m-%d %X", localtime(stat($file)->mtime))),
		src		=> get_cache_filename($self, "$self->{_filename}/", $cached),
		url		=> "$self->{_filename}/$filename",
		height	=> (grep($rotate == $_, (1, 3)) ? $thumbnailwidth  : $thumbnailheight),
		width	=> (grep($rotate == $_, (1, 3)) ? $thumbnailheight : $thumbnailwidth),
	};
	my $comment_ref = get_comment("$file.comment");
	$info->{COMMENT} = $comment_ref->{COMMENT} if $comment_ref->{COMMENT};
	$info->{TITLE} = $comment_ref->{TITLE} if $comment_ref->{TITLE};

	return $info;
} # }}}

sub get_imageinfo { # {{{
	my ($self, $file, $type, $width, $height) = @_;
	my $imageinfo = {
		width	=> $width,
		height	=> $height,
	};
	my $full_imageinfo;

	if ($type eq 'Data stream is not a known image file format') {
		# should never be reached, this is supposed to be handled outside of here
		warn ("Something was fishy with the type of the file $file\n");
	} else { 
		# Some files, like TIFF, PNG, GIF do not have EXIF info embedded but use .thm files
		# instead.
		my $tmpfilename = $file;
		# We have a problem with Windows based file extensions here as they are often .THM
		$tmpfilename =~ s/\.(\w+)$/.thm/;
		if (-e $tmpfilename && -f $tmpfilename && -r $tmpfilename) {
			$full_imageinfo = image_info($tmpfilename);
		}
		# If there is no .thm file and our file is a JPEG file we try to extract the EXIf
		# info using Image::Info
		elsif (grep $type eq $_, qw(JPG)) {
			# Only for files that natively keep the EXIF info in the same file
			$full_imageinfo = image_info($file);
		}
	}

	my @infos = split /\s*,\s*/m, ($self->config('GalleryInfo') || 'Picture Taken => DateTimeOriginal, Flash => Flash');
	foreach (@infos) {

		my ($human_key, $exif_key) = (split /\s*=>\s*/m)[0,1];

		if (defined($exif_key) && defined($full_imageinfo->{$exif_key})) {
			my $value = "";
			if (ref($full_imageinfo->{$exif_key}) eq 'Image::TIFF::Rational') { 
				$value = $full_imageinfo->{$exif_key}->as_string;
			} 
			elsif (ref($full_imageinfo->{$exif_key}) eq 'ARRAY') { # {{{
				foreach my $element (@{$full_imageinfo->{$exif_key}}) {
					if (ref($element) eq 'ARRAY') {
						foreach (@{$element}) {
							$value .= $_ . ' ';
						}
					} 
					elsif (ref($element) eq 'HASH') {
						$value .= "<br>{ ";
			    		foreach (sort keys %{$element}) {
							$value .= "$_ = " . $element->{$_} . ' ';
						}
			    		$value .= "} ";
					} 
					else {
						$value .= $element;
					}
					$value .= ' ';
				}
			} # }}}
			else { # {{{
				my $exif_value = $full_imageinfo->{$exif_key};
				if ($human_key eq 'Flash' && $exif_value =~ m/\d/) {
					my %flashmodes = (
						"0"  => "No",
						"1"  => "Yes",
						"9"  => "Yes",
						"16" => "No (Compulsory) Should be External Flash",
						"17" => "Yes (External)",
						"24" => "No",
						"25" => "Yes (Auto)",
						"73" => "Yes (Compulsory, Red Eye Reducing)",
						"89" => "Yes (Auto, Red Eye Reducing)"
					);
					$exif_value = defined $flashmodes{$exif_value} ? $flashmodes{$exif_value} : 'unknown flash mode';
				}
				$value = $exif_value;
			} # }}}
			if ($exif_key eq 'MeteringMode') { # {{{
				my $exif_value = $full_imageinfo->{$exif_key};
				if ($exif_value =~ /^\d+$/) {
					my %meteringmodes = (
						'0' => 'unknown',
						'1' => 'Average',
						'2' => 'CenterWeightedAverage',
						'3' => 'Spot',
						'4' => 'MultiSpot',
						'5' => 'Pattern',
						'6' => 'Partial',
						'255' => 'Other'
					);
					$exif_value = defined $meteringmodes{$exif_value} ? $meteringmodes{$exif_value} : 'unknown metering mode';
				}
				$value = $exif_value;
				
			} # }}}
			elsif ($exif_key eq 'LightSource') { # {{{
				my $exif_value = $full_imageinfo->{$exif_key};
				if ($exif_value =~ /^\d+$/) {
					my %lightsources = (
						'0' => 'unknown',
						'1' => 'Daylight',
						'2' => 'Fluorescent',
						'3' => 'Tungsten (incandescent light)',
						'4' => 'Flash',
						'9' => 'Fine weather',
						'10' => 'Cloudy weather',
						'11' => 'Shade',
						'12' => 'Daylight fluorescent',
						'13' => 'Day white fluorescent',
						'14' => 'Cool white fluorescent',
						'15' => 'White fluorescent',
						'17' => 'Standard light A',
						'18' => 'Standard light B',
						'19' => 'Standard light C',
						'20' => 'D55',
						'21' => 'D65',
						'22' => 'D75',
						'23' => 'D50',
						'24' => 'ISO studio tungsten',
						'255' => 'other light source'
					);
					$exif_value = defined $lightsources{$exif_value} ? $lightsources{$exif_value} : 'unknown light source';
				}
				$value = $exif_value;
			} # }}}
			elsif ($exif_key eq 'FocalLength') { # {{{
				if ($value =~ /^(\d+)\/(\d+)$/) {
					$value = eval { $1 / $2 };
					if ($@) {
						$value = $@;
					} else {
						$value = int($value + 0.5) . "mm";

					}
				}
			} # }}}
			elsif ($exif_key eq 'ShutterSpeedValue') { # {{{
				if ($value =~ /^((?:\-)?\d+)\/(\d+)$/) {
					$value = eval { $1 / $2 };
					if ($@) {
						$value = $@;
					} else {
						eval {
							$value = 1/(exp($value*log(2)));
							if ($value < 1) {
								$value = "1/" . (int((1/$value)));
							} else {
						  	 	$value = int($value*10)/10; 
							}
						};
						if ($@) {
							$value = $@;
						} else {
							$value = $value . " sec";
						}
					}
				}
			} # }}}
			elsif ($exif_key eq 'ApertureValue') { # {{{
				if ($value =~ /^(\d+)\/(\d+)$/) {
					$value = eval { $1 / $2 };
					if ($@) {
						$value = $@;
					} else {
						# poor man's rounding
						$value = int(exp($value*log(2)*0.5)*10)/10;
						$value = "f" . $value;
					}
				}
			} # }}}
			elsif ($exif_key eq 'FNumber') { # {{{
				if ($value =~ /^(\d+)\/(\d+)$/) {
					$value = eval { $1 / $2 };
					if ($@) {
						$value = $@;
					} else {
						$value = int($value*10+0.5)/10;
						$value = "f" . $value;
					}
				}
			} # }}}
			$imageinfo->{$human_key} = $value;
		} 
	}

	if ($self->config('GalleryUseFileDate') &&
		($self->config('GalleryUseFileDate') eq '1'
		|| !$imageinfo->{"Picture Taken"} )) {

		my $st = stat($file);
		$imageinfo->{"DateTimeOriginal"} = $imageinfo->{"Picture Taken"} = scalar localtime($st->mtime) if $st;
	}

	return $imageinfo;
} # }}}

sub readfile_getnum { # {{{
	my ($self, $imageinfo, $filename) = @_;

	my $rotate = 0;

	# Check to see if the image contains the Orientation EXIF key,
	# but allow user to override using rotate
	if (!defined($self->config("GalleryAutoRotate")) 
		|| $self->config("GalleryAutoRotate") eq "1") {
		if (defined($imageinfo->{Orientation})) {
			if ($imageinfo->{Orientation} eq m/right_top/i) {
				$rotate = 1;
			}	
			elsif ($imageinfo->{Orientation} =~ m/left_bot/i) {
				$rotate = 3;
			}
		}
	}

	if (open(FH, "<$filename")) {
		my $temp = <FH>;
		close(FH);
		$temp = int($temp || 0);
		unless ($temp == 1 || $temp == 2 || $temp == 3) {
			$rotate = 0;
		}
		$rotate = $temp;
	}

	return $rotate;
} # }}}

sub get_filecontent { # {{{
	my ($self, $file) = @_;
	open(FH, $file) or return undef;
	my $content = '';
	{
		local $/;
		$content = <FH>;
	}
	close(FH);
	return $content;
} # }}}

sub get_dirtitle($$;$) { # {{{
	my ($self, $dirname, $title) = @_;
	if (defined $self->{_dirtitle_cache}
			and defined $self->{_dirtitle_cache}->{$dirname}) {

		return $self->{_dirtitle_cache}->{$dirname};
	}
	else {
		my $dirtitle = get_filecontent($self, "$dirname/.folder");
		$dirtitle ||= $title || '';

		$dirtitle =~ s/_/ /g if $self->config('GalleryUnderscoresToSpaces');
		$dirtitle =~ s/\s+/ /g;
		$dirtitle =~ s/^\s+//;
		$dirtitle =~ s/\s+$//;

		$self->{_dirtitle_cache}->{$dirname} = $dirtitle;

		return $self->{_dirtitle_cache}->{$dirname};
	}
} # }}}

sub get_comment { # {{{
	my $filename = shift;
	my $comment_ref = {
		TITLE	=> undef,
		COMMENT	=> '',
	};

	open(FH, $filename) or return $comment_ref;
	my $title = <FH>;
	if ($title =~ m/^TITLE: (.*)$/) {
		chomp($comment_ref->{TITLE} = $1);
	} 
	else {
		$comment_ref->{COMMENT} = $title;
	}

	{
		local $/;
		$comment_ref->{COMMENT} .= <FH>;
	}
	close(FH);

	return $comment_ref;
} # }}}

sub scale_picture { # {{{

	my ($self, $fullpath, $width, $height, $imageinfo) = @_;

	my @dirs = split(/\//, $fullpath);
	my $filename = pop(@dirs);

	my ($orig_width, $orig_height, $type) = imgsize($fullpath);

	my $cache = cache_dir($self, 1);

	my $newfilename = get_scaled_picture_name($self, $fullpath, $type, $width, $height);

	if (($width > $orig_width) && ($height > $orig_height)) {
		File::Copy::copy($fullpath, "$cache/$newfilename") or warn "Can't copy file: $!";
		return $newfilename;
	}

	my ($thumbnailwidth, $thumbnailheight) = get_thumbnailsize($self, $orig_width, $orig_height);

	# Do we want to generate a new file in the cache?
	my $scale = 1;

	my $newpath = "$cache/$newfilename";
	if (-f $newpath) {	
		$scale = 0;

		# Check to see if the image has changed
		my $filestat = stat($fullpath);
		my $cachestat = stat($newpath);
		if ($filestat->mtime > $cachestat->mtime) {
			$scale = 1;
		}	

		# Check to see if the .rotate file has been added or changed
		if (-f $fullpath . ".rotate") {
			my $rotatestat = stat($fullpath . ".rotate");
			if ($rotatestat->mtime > $cachestat->mtime) {
				$scale = 1;
			}	
		}		
		# Check to see if the copyrightimage has been added or changed
		if ($self->config('GalleryCopyrightImage') && -f $self->config('GalleryCopyrightImage')) {
			unless ($width == $thumbnailwidth or $width == $thumbnailheight) {
				my $copyrightstat = stat($self->config('GalleryCopyrightImage'));
				if ($copyrightstat->mtime > $cachestat->mtime) {
					$scale = 1;
				}	
			}
		}	

	}	

	if ($scale) {

		my $rotate = readfile_getnum($self, $imageinfo, $fullpath . ".rotate");
		my $quality = int($self->config('GalleryQuality') || 85) || 85;

		if ($width == $thumbnailwidth or $width == $thumbnailheight) {

			resizepicture($self, $fullpath, $newpath, $width, $height, $rotate, '', '', '', '', '');

		} else {

			resizepicture($self, $fullpath, $newpath, $width, $height, $rotate, 
				($self->config('GalleryCopyrightImage')	|| ''), 
				($self->config('GalleryTTFDir')			|| ''), 
				($self->config('GalleryCopyrightText')	|| ''), 
				($self->config('GalleryCopyrightColor')	|| ''), 
				($self->config('GalleryTTFFile')		|| ''), 
				($self->config('GalleryTTFSize')		|| ''),
				$quality);

		}
	}

	return $newfilename;

} # }}}

sub generate_menu { # {{{

	my ($self) = @_;

	my $root_text = ($self->config('GalleryRootText') || "root:");
	my $root_path = ($self->config('GalleryRootURI') || "/");

	my $filename = $self->{_filename};
	my $uri = $self->{_uri};
	$uri =~ s#^$root_path#/#g;

	my @links = grep { $_ = '' unless defined $_ } split (/\//, $uri);

	my $picturename;
	if (-f $filename) {
		$picturename = pop(@links);	
	}

	if ($uri eq $root_path) {
		return [ {
			'link'	=> $root_path,
			'desc'	=> $root_text,
		}];
	}

	my @menu;
	my $menuurl = '';

	unless (length($links[0])) {
		shift @links;
		push @menu, {
			'link'	=> $root_path,
			'desc'	=> $root_text,
		};
	}

	my $dirname = '';
	foreach my $link (@links) {

		my $linktext = $link;
		$menuurl .= "/$link";

		$dirname = File::Spec->catdir($dirname, $link);

		if (-e $dirname . ".folder") {
			$linktext = get_filecontent($self, $dirname . ".folder");
		}

		if ($uri eq $menuurl) {
			push @menu, {
				'desc'	=> $linktext,
			};
		}
		else {
			push @menu, {
				'link'	=> "$menuurl/",
				'desc'	=> $linktext,
			};
		}

	}

	if (-f $filename) {
		push @menu, {
			'desc'	=> $picturename,
		};
	}

	return \@menu;
} # }}}

sub resizepicture { # {{{
	my ($self, $infile, $outfile, $x, $y, $rotate, $copyrightfile, $GalleryTTFDir, $GalleryCopyrightText, $text_color, $GalleryTTFFile, $GalleryTTFSize, $quality) = @_;

	# Load image
	my $image = Image::Imlib2->load($infile) or warn("Unable to open file $infile, $!");

	# Scale image
	$image = $image->create_scaled_image($x, $y) or warn("Unable to scale image $infile. Are you running out of memory?");

	# Rotate image
	if ($rotate != 0) {
		$image->image_orientate($rotate);
	}

	# blend copyright image onto image
 	if ($copyrightfile ne '') {
		if (-f $copyrightfile and (my $logo=Image::Imlib2->load($copyrightfile))) {
			$x = $image->get_width();
			$y = $image->get_height();
			my $logox = $logo->get_width();
			my $logoy = $logo->get_height();
			$image->blend($logo, 0, 0, 0, $logox, $logoy, $x-$logox, $y-$logoy, $logox, $logoy);
		}
		else {
			warn ("GalleryCopyrightImage '$copyrightfile' was not found\n");
		}
	}

	if ($GalleryTTFDir && $GalleryCopyrightText && $GalleryTTFFile && $text_color) {
		if (!-d $GalleryTTFDir) {

			warn ("GalleryTTFDir $GalleryTTFDir is not a dir\n");

		} elsif ($GalleryCopyrightText eq '') {

			warn ("GalleryCopyrightText is empty. No text inserted to picture\n");

		} elsif (!-e "$GalleryTTFDir/$GalleryTTFFile") {

			warn ("GalleryTTFFile $GalleryTTFFile was not found\n");

		} else {
 
			$GalleryTTFFile =~ s/\.TTF$//i;
			$image->add_font_path("$GalleryTTFDir");

			my ($r_val, $g_val, $b_val, $a_val) = split (/,/, $text_color);

			$image->set_colour($r_val, $g_val, $b_val, $a_val);

			$image->load_font("$GalleryTTFFile/$GalleryTTFSize");
			my($text_x, $text_y) = $image->get_text_size("$GalleryCopyrightText");
			$x = $image->get_width();
			$y = $image->get_height();

			my $offset = 3;

			if (($text_x < $x - $offset) && ($text_y < $y - $offset)) {
				$image->draw_text($x-$text_x-$offset, $y-$text_y-$offset, "$GalleryCopyrightText");
			} else {
				warn ("Text is to big for the picture.\n");
			}
		}
	}

	if ($quality && $quality =~ m/^\d+$/) {
		$image->set_quality($quality);
	}

	$image->save($outfile);

} # }}}

sub gallerysort($$;) { # {{{
	my ($self, $filename, @files) = @_;

	my $sortby		= $self->config('GallerySortBy') || '';
	$filename = (File::Spec->splitpath($filename))[1] if (-f $filename);
	if ($sortby && $sortby =~ m/^(size|atime|mtime|ctime)$/) {
		@files = map(/^\d+ (.*)/, sort map(stat("$filename/$_")->$sortby()." $_", @files));
	} else {
		@files = sort @files;
	}
	return @files;
} # }}}

sub get_subentries_count($$;)
{ # {{{
	my ($self, $topdir, $dir) = @_;
	my $data = {
		files	=> 0,
		dirs	=> 0,
	};

	my $doc_pattern		= ($self->config('GalleryDocFile') or '\.(mpe?g|avi|mov|asf|wmv|doc|mp3|ogg|pdf|rtf|wav|dlt|html?|csv|eps|sxw)$');
	my $img_pattern		= ($self->config('GalleryImgFile') or '\.(jpe?g|png|tiff?|ppm)$');

	opendir(DIR, "$topdir/$dir") or return $data;
	my @files	= grep { !/^\./ && !/^CVS$/ && -f "$topdir/$dir/$_" } readdir (DIR);
	rewinddir(DIR);
	my @dirs	= grep { !/^\./ && !/^CVS$/ && -d "$topdir/$dir/$_" } readdir (DIR);
	closedir(DIR);

	my @new_files = ();
	foreach my $picture (@files) {
		if ($picture =~ /$doc_pattern/i or $picture =~ /$img_pattern/i) {
			push (@new_files, $picture);
		}
	}
	@files = @new_files;

	$data->{files}	= scalar(@files);
	$data->{dirs}	= scalar(@dirs);

	return $data;

} # }}}

1;

# Documentation {{{

=head1 NAME

Apache::Gallery - mod_perl handler to create an image gallery

=head1 SYNOPSIS

See the INSTALL file in the distribution for installation instructions.

=head1 DESCRIPTION

Apache::Gallery creates an thumbnail index of each directory and allows 
viewing pictures in different resolutions. Pictures are resized on the 
fly and cached. The gallery can be configured and customized in many ways
and a custom copyright image can be added to all the images without
modifying the original.

=head1 CONFIGURATION

In your httpd.conf you set the global options for the gallery. You can
also override each of the options in .htaccess files in your gallery
directories.

The options are set in the httpd.conf/.htaccess file using the syntax:
B<PerlSetVar OptionName 'value'>

Example: B<PerlSetVar GalleryCacheDir '/var/tmp/Apache-Gallery/'>

=over 4

=item B<GalleryAutoRotate>

Some cameras, like the Canon G3, can detect the orientation of a 
the pictures you take and will save this information in the 
'Orientation' EXIF field. Apache::Gallery will then automatically
rotate your images. 

This behavior is default but can be disabled by setting GalleryAutoRotate
to 0.

=item B<GalleryCacheDir>

Directory where Apache::Gallery should create its cache with scaled
pictures. The default is /var/tmp/Apache-Gallery/ . Here, a directory
for each virtualhost or location will be created automaticly. Make
sure your webserver has write access to the CacheDir.

=item B<GalleryTemplateDir>

Full path to the directory where you placed the templates. This option
can be used both in your global configuration and in .htaccess files,
this way you can have different layouts in different parts of your 
gallery.

No default value, this option is required.

=item B<GalleryInfo>

With this option you can define which EXIF information you would like
to present from the image. The format is: '<MyName => KeyInEXIF, 
MyOtherName => OtherKeyInEXIF'

Examples of keys: B<ShutterSpeedValue>, B<ApertureValue>, B<SubjectDistance>,
and B<Camera>

You can view all the keys from the EXIF header using this perl-oneliner:

perl C<-e> 'use Data::Dumper; use Image::Info qw(image_info); print Dumper(image_info(shift));' filename.jpg

Default is: 'Picture Taken => DateTimeOriginal, Flash => Flash'

=item B<GallerySizes>

Defines which widths images can be scaled to. Images cannot be
scaled to other widths than the ones you define with this option.

The default is '640 800 1024 1600'

=item B<GalleryThumbnailSize>

Defines the width and height of the thumbnail images. 

Defaults to '100x75'

=item B<GalleryThumbnailSizeLS>

If set to '1', B<GalleryThumbnailSize> is the long and the short side of
the thumbnail image instead of the width and height.

Defaults to '0'.

=item B<GalleryCopyrightImage>

Image you want to blend into your images in the lower right
corner. This could be a transparent png saying "copyright
my name 2001".

Optional.

=item B<GalleryWrapNavigation>

Make the navigation in the picture view wrap around (So Next
at the end displays the first picture, etc.)

Set to 1 or 0, default is 0

=item B<GalleryAllowOriginal>

Allow the user to download the Original picture without
resizing or putting the CopyrightImage on it.

Set to 1 or 0, default is 0

=item B<GallerySlideshowIntervals>

With this option you can configure which intervals can be selected for
a slideshow. The default is '3 5 10 15 30'

=item B<GallerySortBy>

Instead of the default filename ordering you can sort by any
stat attribute. For example size, atime, mtime, ctime.

=item B<GalleryDirSortBy>

Set this variable to sort directories differently than other items,
can be set to size, atime, mtime and ctime; setting any other value
will revert to sorting by name.

=item B<GalleryMemoize>

Cache EXIF data using Memoize - this will make Apache::Gallery faster
when many people access the same images, but it will also cache EXIF
data until the current Apache child dies.

=item B<GalleryUseFileDate>

Set this option to 1 to make A::G show the files timestamp
instead of the EXIF value for "Picture taken".

=item B<GallerySelectionMode>

Enable the selection mode. Select images with checkboxes and
get a list of filenames. 

=item B<GalleryEXIFMode>

You can choose how Apache::Gallery should display EXIF info
from your images. 

The default setting is 'namevalue'. This setting will make 
Apache::Gallery print out the names and values of the EXIF values 
you configure with GalleryInfo. The information will be parsed into 
$INFO in pictureinfo.tpl.  

You can also set it to 'values' which will make A::G parse
the configured values into the var $EXIFVALUES as 'value | value | value'

If you set this option to 'variables' the items you configure in GalleryInfo 
will be available to your templates as $EXIF_<KEYNAME> (in all uppercase). 
That means that with the default setting "Picture Taken => DateTimeOriginal, 
Flash => Flash" you will have the variables $EXIF_DATETIMEORIGINAL and 
$EXIF_FLASH avilable to your templates. You can place them
anywhere you want.

=item B<GalleryRootPath>

Change the location of gallery root. The default is ""

=item B<GalleryRootText>

Change the name that appears as the root element in the menu. The
default is "root:"

=item B<GalleryMaxThumbnailsPerPage>

This options controls how many thumbnails should be displayed in a 
page. It requires $BROWSELINKS to be in the index.tpl template file.

=item B<GalleryImgFile>

Pattern matching the files you want Apache::Gallery to view in the
index as thumbnails. 

The default is '\.(jpe?g|png|tiff?|ppm)$'

=item B<GalleryDocFile>

Pattern matching the files you want Apache::Gallery to view in the index
as normal files. All other filetypes will still be served by Apache::Gallery
but are not visible in the index.

The default is '\.(mpe?g|avi|mov|asf|wmv|doc|mp3|ogg|pdf|rtf|wav|dlt|html?|csv|eps|sxw)$'

=item B<GalleryTTFDir>

To use the GalleryCopyrightText feature you must set this option to the
directory where your True Type fonts are stored. No default is set.

Example:

	PerlSetVar      GalleryTTFDir '/usr/share/fonts/'

=item B<GalleryTTFFile>

To use the GalleryCopyrightText feature this option must be set to the
name of the True Type font you wish to use. Example:

	PerlSetVar      GalleryTTFFile 'verdanab.ttf'

=item B<GalleryTTFSize>

Configure the size of the CopyrightText that will be inserted as 
copyright notice in the corner of your pictures.

Example:

	PerlSetVar      GalleryTTFSize '10'

=item B<GalleryCopyrightText>

The text that will be inserted as copyright notice.

Example:

        PerlSetVar      GalleryCopyrightText '(c) Michael Legart'

=item B<GalleryCopyrightColor>

The text color of your copyright notice.

Examples:

White:
        PerlSetVar      GalleryCopyrightColor '255,255,255,255'

Black:
        PerlSetVar      GalleryCopyrightColor '0,0,0,255'

Red:
        PerlSetVar      GalleryCopyrightColor '255,0,0,255'

Green:
        PerlSetVar      GalleryCopyrightColor '0,255,0,255'

Blue:
        PerlSetVar      GalleryCopyrightColor '0,0,255,255'

Transparent orange:
        PerlSetVar      GalleryCopyrightColor '255,127,0,127'

=item B<GalleryQuality>

The quality (1-100) of scaled images

This setting affects the quality of the scaled images.
Set this to a low number to reduce the size of the scaled images.
Remember to clear out your cache if you change this setting.
Quality seems to default to 75, at least in the jpeg and png loader code in
Imlib2 1.1.0.

Examples:

Quality at 50:
        PerlSetVar      GalleryQuality '50'

=item B<GalleryUnderscoresToSpaces>

Set this option to 1 to convert underscores to spaces in the listing
of directory names.

=back

=head1 FEATURES

=over 4

=item B<Rotate images>

Some cameras, like the Canon G3, detects the orientation of a picture
and adds this info to the EXIF header. Apache::Gallery detects this
and automaticly rotates images with this info.

If your camera does not support this, you can rotate the images 
manually, This can also be used to override the rotate information
from a camera that supports that. You can also disable this behavior
with the GalleryAutoRotate option.

To use this functionality you have to create file with the name of the 
picture you want rotated appened with ".rotate". The file should include 
a number where these numbers are supported:

	"1", rotates clockwise by 90 degree
	"2", rotates clockwise by 180 degrees
	"3", rotates clockwise by 270 degrees

So if we want to rotate "Picture1234.jpg" 90 degrees clockwise we would
create a file in the same directory called "Picture1234.jpg.rotate" with
the number 1 inside of it.

=item B<Comments>

To include comments for a directory you create a <directory>.comment
file where the first line can contain "TITLE: New title" which
will be the title of the page, and a comment on the following 
lines.
To include comments for each picture you create files called 
picture.jpg.comment where the first line can contain "TITLE: New
title" which will be the title of the page, and a comment on the
following lines.

Example:

	TITLE: This is the new title of the page
	And this is the comment.<br>
	And this is line two of the comment.

The visible name of the folder is by default identical to the name of
the folder, but can be changed by creating a file <directory>.folder
with the visible name of the folder.

=back

=head1 DEPENDENCIES

=over 4

=item B<Perl 5>

=item B<Apache with mod_perl>

=item B<URI::Escape>

=item B<Image::Info>

=item B<Image::Size>

=item B<Text::Template>

=item B<Image::Imlib2>

=item B<X11 libraries>
(ie, XFree86)

=item B<Imlib2>
Remember the -dev package when using rpm, deb or other package formats!

=back

=head1 AUTHOR

Michael Legart <michael@legart.dk>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2001-2004 Michael Legart <michael@legart.dk>

Templates designed by Thomas Kjaer <tk@lnx.dk>

Apache::Gallery is free software and is released under the Artistic License.
See B<http://www.perl.com/language/misc/Artistic.html> for details.

The video icons are from the GNOME project. B<http://www.gnome.org/>

=head1 THANKS

Thanks to Thomas Kjaer for templates and design of B<http://apachegallery.dk>
Thanks to Thomas Eibner and other for patches. (See the Changes file)

=head1 SEE ALSO

L<perl>, L<mod_perl>, L<Image::Imlib2>, L<CGI::FastTemplate>,
L<Image::Info>, and L<Image::Size>.

=cut

# }}}

# vim600: fdm=marker fdl=0 fdc=3

